package ardash.doodledudette.canvas;

import java.util.ArrayDeque;
import java.util.Queue;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.InputAdapter;
import com.badlogic.gdx.InputProcessor;
import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.graphics.Pixmap;
import com.badlogic.gdx.graphics.Texture;
import com.badlogic.gdx.graphics.TextureData;
import com.badlogic.gdx.graphics.g2d.SpriteBatch;
import com.badlogic.gdx.utils.Disposable;
import com.badlogic.gdx.utils.Pool;
import com.badlogic.gdx.utils.Pools;
import com.github.czyzby.kiwi.util.gdx.asset.Disposables;

import ardash.doodledudette.HSLColor;
import ardash.doodledudette.SVGAssistant;

public class CanvasAggregate implements Disposable {
	private static final String TAG = "doodledudette.CanvasAggregate";
	private SpriteBatch batch;
	private InputProcessor inputProcessor;
	private Pixmap canvasPixmap;
	private Texture canvasTexture;
	private Texture svgTexture;
	private int canvasSize;
	private boolean[][] blackPixels;
	private int[][] disjunctArea;
	private int currentArea;
	private HSLColor currentColor;
	private int radius;
	private boolean dragging = false;

	public CanvasAggregate() {
		currentColor = new HSLColor(0f, 1f, 0.5f);
	}

	public void findDisjunctAreas() {
		Gdx.app.log(TAG, "start findDisjunctAreas");
		int areaCount = 0;
		final int cs = getCanvasSize();
		// Iterate over all pixels
		for (int y = 0; y < cs; y++) {
			for (int x = 0; x < cs; x++) {
				// If pixel is white and not yet visited, start flood fill
				if (!getBlackPixels()[x][y] && getDisjunctArea()[x][y] == 0) {
					areaCount++;
					floodFill(x, y, areaCount);
				}
			}
		}
		Gdx.app.log(TAG, "found " + areaCount + " areas");
		Gdx.app.log(TAG, "end    findDisjunctAreas");
	}

	private void floodFill(int x, int y, int areaCount) {
		final Pool<CheapPoint> pool = Pools.get(CheapPoint.class, 1000);
		Queue<CheapPoint> queue = new ArrayDeque<>();
		final CheapPoint firstPoint = pool.obtain();
		firstPoint.x = x;
		firstPoint.y = y;
		queue.add(firstPoint);
		disjunctArea[x][y] = areaCount;

		while (!queue.isEmpty()) {
			CheapPoint current = queue.poll();
			int cx = current.x;
			int cy = current.y;

			// Check and add adjacent pixels if they are white and not yet visited
			if (isValidPixel(cx + 1, cy) && !blackPixels[cx + 1][cy] && disjunctArea[cx + 1][cy] == 0) {
				final CheapPoint toBeQueued = pool.obtain();
				toBeQueued.x = cx + 1;
				toBeQueued.y = cy;
				queue.add(toBeQueued);
				disjunctArea[cx + 1][cy] = areaCount;
			}
			if (isValidPixel(cx - 1, cy) && !blackPixels[cx - 1][cy] && disjunctArea[cx - 1][cy] == 0) {
				final CheapPoint toBeQueued = pool.obtain();
				toBeQueued.x = cx - 1;
				toBeQueued.y = cy;
				queue.add(toBeQueued);
				disjunctArea[cx - 1][cy] = areaCount;
			}
			if (isValidPixel(cx, cy + 1) && !blackPixels[cx][cy + 1] && disjunctArea[cx][cy + 1] == 0) {
				final CheapPoint toBeQueued = pool.obtain();
				toBeQueued.x = cx;
				toBeQueued.y = cy + 1;
				queue.add(toBeQueued);
				disjunctArea[cx][cy + 1] = areaCount;
			}
			if (isValidPixel(cx, cy - 1) && !blackPixels[cx][cy - 1] && disjunctArea[cx][cy - 1] == 0) {
				final CheapPoint toBeQueued = pool.obtain();
				toBeQueued.x = cx;
				toBeQueued.y = cy - 1;
				queue.add(toBeQueued);
				disjunctArea[cx][cy - 1] = areaCount;
			}
			pool.free(current);
		}
	}

	private boolean isValidPixel(int x, int y) {
		return x >= 0 && x < getCanvasSize() && y >= 0 && y < getCanvasSize();
	}

	private boolean isInsideDisjunctArea(int x, int y) {
		if (isValidPixel(x, y)) {
			int areaNumber = getDisjunctArea()[x][y];
			return areaNumber != 0 && areaNumber == getCurrentArea();
		}
		return false;
	}

	public void drawCircle(float x, float y, boolean dragged) {
		// Check if the touch position is within the identified disjunct area
		if (isInsideDisjunctArea((int) x, (int) y) || dragged) {

			// Iterate over the pixels within the circle's bounding box
			for (int px = (int) x - radius; px <= (int) x + radius; px++) {
				for (int py = (int) y - radius; py <= (int) y + radius; py++) {
					// Check if the current pixel is within the circle's radius
					if ((px - x) * (px - x) + (py - y) * (py - y) <= radius * radius) {
						// Check if the pixel is within the bounds of the canvas
						// TODO large circles are slower, so slow devices with high res displays are
						// foobared
						// TODO optimize performance, use variables instead of accessors
						// TODO use isvalidpixel
						if (px >= 0 && px < getCanvasSize() && py >= 0 && py < getCanvasSize()) {
							if (!getBlackPixels()[px][py] && isInsideDisjunctArea(px, py)) {
								getCanvasPixmap().setColor(currentColor.getRGBColor());
								getCanvasPixmap().drawPixel(px, py);

							} else if (getBlackPixels()[px][py]) {
								// TODO thresholding not working, so redraw semi-transparant pixels in black =
								// slower
								getCanvasPixmap().setColor(Color.BLACK);
								getCanvasPixmap().drawPixel(px, py);

							}
						}
					}
				}
			}

			// sometimes disposal is too fast when leaving the screen and the
			// InputEventQueue.drain() is still
			// running. The this will cause an exception and close the game. Bugfix for #14
			if (!getCanvasPixmap().isDisposed()) {
				// Update the canvas texture
				getCanvasTexture().draw(getCanvasPixmap(), 0, 0);
			}
		}
	}

	public void createInputProcessor() {
		setInputProcessor(new InputAdapter() {
			@Override
			public boolean touchDown(int screenX, int screenY, int pointer, int button) {

				// do not handle this event if outside of touch area
				if (screenX >= canvasSize) {
					return false;
				}

				dragging = true;

				setCurrentArea(getDisjunctArea()[screenX][(screenY)]);
				System.out.println("currentArea: " + getCurrentArea());
				// Draw a circle at touch location
				drawCircle(screenX, screenY, false);
//	                }
				return true;
			}

			@Override
			public boolean touchDragged(int screenX, int screenY, int pointer) {
				drawCircle(screenX, screenY, true);
				// TODO if the circles must also be drawn on a line between the circles (if the
				// finger moves very fast) the method drawCircle can be called repeatedly here
				return true;
			}

			@Override
			public boolean touchUp(int screenX, int screenY, int pointer, int button) {
				dragging = false;
				return super.touchUp(screenX, screenY, pointer, button);
			}
		});
	}

	public void buildCanvas(String svgzFile) {
		Gdx.app.log(TAG, "start buildCanvas");
		setBatch(new SpriteBatch());

		// Calculate the size of the canvas
		setCanvasSize(Gdx.graphics.getHeight());

		// Initialize the canvas Pixmap and Texture
		final int cs = getCanvasSize();
		setCanvasPixmap(new Pixmap(cs, cs, Pixmap.Format.RGBA8888));
		final Pixmap cp = getCanvasPixmap();
		cp.setColor(Color.WHITE);
		cp.fill();

		// Load the clipart image
		Gdx.app.log(TAG, "loadGzippedSVGTexture ...");
		Texture clipartTexture = SVGAssistant.loadGzippedSVGTexture(svgzFile, cs, true);

		TextureData textureData = clipartTexture.getTextureData();
		if (!textureData.isPrepared()) {
			Gdx.app.log(TAG, "textureData.prepare() ...");
			textureData.prepare();
		}
		Gdx.app.log(TAG, "textureData.consumePixmap() ...");
		Pixmap clipartPixmap = textureData.consumePixmap();

		// Scale and draw the clipart onto the canvas
		Gdx.app.log(TAG, "getCanvasPixmap().drawPixmap() ...");
		cp.drawPixmap(clipartPixmap, 0, 0, clipartPixmap.getWidth(), clipartPixmap.getHeight(), // source area
				0, 0, cs, cs); // destination area

		// Track black pixels
		Gdx.app.log(TAG, "setBlackPixels() ...");
		setBlackPixels(new boolean[cs][cs]);
		// optimized, reused Color instances, took almost 2 seconds on ARM
		final Color color = new Color();
		final int intBitsWhite = Color.WHITE.toIntBits();
		Gdx.app.log(TAG, "setBlackPixels() ...");
		for (int y = 0; y < cs; y++) {
			for (int x = 0; x < cs; x++) {
				int pixel = cp.getPixel(x, y);
				Color.rgba8888ToColor(color, pixel);
				if (color.toIntBits() != intBitsWhite) {
					getBlackPixels()[x][y] = true;
				}
			}
		}

		// Identify disjunct areas
		setDisjunctArea(new int[cs][cs]);
		Gdx.app.log(TAG, "findDisjunctAreas() ...");
		// TODO optimize, reuse Color instances takes almost 3 seconds on ARM
		findDisjunctAreas();

		// Dispose the clipart texture and pixmap
		Gdx.app.log(TAG, "disposing ...");
		clipartTexture.dispose();
		clipartPixmap.dispose();

		// Create the canvas texture
		Gdx.app.log(TAG, "setCanvasTexture() ...");
		setCanvasTexture(new Texture(cp));
		Gdx.app.log(TAG, "end   buildCanvas");
	}

	public void renderCanvasTexture() {
		getBatch().begin();
		getBatch().draw(getCanvasTexture(), 0, 0, getCanvasSize(), getCanvasSize());
		getBatch().end();
	}

	public SpriteBatch getBatch() {
		return batch;
	}

	public void setBatch(SpriteBatch batch) {
		this.batch = batch;
	}

	public InputProcessor getInputProcessor() {
		return inputProcessor;
	}

	public void setInputProcessor(InputProcessor inputProcessor) {
		this.inputProcessor = inputProcessor;
	}

	public Pixmap getCanvasPixmap() {
		return canvasPixmap;
	}

	public void setCanvasPixmap(Pixmap canvasPixmap) {
		this.canvasPixmap = canvasPixmap;
	}

	public Texture getCanvasTexture() {
		return canvasTexture;
	}

	public void setCanvasTexture(Texture canvasTexture) {
		this.canvasTexture = canvasTexture;
	}

	public Texture getSvgTexture() {
		return svgTexture;
	}

	public void setSvgTexture(Texture svgTexture) {
		this.svgTexture = svgTexture;
	}

	public int getCanvasSize() {
		return canvasSize;
	}

	public void setCanvasSize(int canvasSize) {
		this.canvasSize = canvasSize;
	}

	public boolean[][] getBlackPixels() {
		return blackPixels;
	}

	public void setBlackPixels(boolean[][] blackPixels) {
		this.blackPixels = blackPixels;
	}

	public int[][] getDisjunctArea() {
		return disjunctArea;
	}

	public void setDisjunctArea(int[][] disjunctArea) {
		this.disjunctArea = disjunctArea;
	}

	public int getCurrentArea() {
		return currentArea;
	}

	public void setCurrentArea(int currentArea) {
		this.currentArea = currentArea;
	}

	public HSLColor getCurrentColor() {
		return currentColor;
	}

	public void setCurrentColor(HSLColor currentColor) {
		this.currentColor = currentColor;
	}

	public void setRadius(float radius) {
		this.radius = (int) radius;
	}

	public void setRadius(int radius) {
		this.radius = radius;
	}

	public boolean isDragging() {
		return dragging;
	}

	@Override
	public void dispose() {
		Disposables.gracefullyDisposeOf(batch, canvasPixmap, canvasTexture, svgTexture);
	}
}